from datetime import datetime
from typing import ClassVar

from pydantic import field_validator

from liti.core.model.v1.operation.data.base import Operation
from liti.core.model.v1.schema import ColumnName, ForeignKey, Identifier, IntervalLiteral, PrimaryKey, RoundingMode, \
    Table, QualifiedName


class CreateTable(Operation):
    """ Semantics: CREATE """

    table: Table

    KIND: ClassVar[str] = 'create_table'


class DropTable(Operation):
    """ Semantics: DROP """

    table_name: QualifiedName

    KIND: ClassVar[str] = 'drop_table'


class RenameTable(Operation):
    from_name: QualifiedName
    to_name: Identifier

    KIND: ClassVar[str] = 'rename_table'


class SetPrimaryKey(Operation):
    table_name: QualifiedName
    primary_key: PrimaryKey | None = None

    KIND: ClassVar[str] = 'set_primary_key'


class AddForeignKey(Operation):
    table_name: QualifiedName
    foreign_key: ForeignKey

    KIND: ClassVar[str] = 'add_foreign_key'


class DropConstraint(Operation):
    table_name: QualifiedName
    constraint_name: Identifier

    KIND: ClassVar[str] = 'drop_constraint'


class SetPartitionExpiration(Operation):
    table_name: QualifiedName
    expiration_days: float | None = None

    KIND: ClassVar[str] = 'set_partition_expiration'


class SetRequirePartitionFilter(Operation):
    table_name: QualifiedName
    require_filter: bool

    KIND: ClassVar[str] = 'set_require_partition_filter'


class SetClustering(Operation):
    table_name: QualifiedName
    column_names: list[ColumnName] | None = None

    KIND: ClassVar[str] = 'set_clustering'

    @field_validator('column_names', mode='before')
    @classmethod
    def validate_column_names(cls, value: list[ColumnName] | None) -> list[ColumnName] | None:
        if value:
            return value
        else:
            return None


class SetDescription(Operation):
    table_name: QualifiedName
    description: str | None = None

    KIND: ClassVar[str] = 'set_description'


class SetLabels(Operation):
    table_name: QualifiedName
    labels: dict[str, str] | None = None

    KIND: ClassVar[str] = 'set_labels'


class SetTags(Operation):
    table_name: QualifiedName
    tags: dict[str, str] | None = None

    KIND: ClassVar[str] = 'set_tags'


class SetExpirationTimestamp(Operation):
    table_name: QualifiedName
    expiration_timestamp: datetime | None = None

    KIND: ClassVar[str] = 'set_expiration_timestamp'


class SetDefaultRoundingMode(Operation):
    table_name: QualifiedName
    rounding_mode: RoundingMode | None = None

    KIND: ClassVar[str] = 'set_default_rounding_mode'


class SetMaxStaleness(Operation):
    table_name: QualifiedName
    max_staleness: IntervalLiteral | None = None

    KIND: ClassVar[str] = 'set_max_staleness'


class SetEnableChangeHistory(Operation):
    table_name: QualifiedName
    enabled: bool

    KIND: ClassVar[str] = 'set_enable_change_history'


class SetEnableFineGrainedMutations(Operation):
    table_name: QualifiedName
    enabled: bool

    KIND: ClassVar[str] = 'set_enable_fine_grained_mutations'


class SetKmsKeyName(Operation):
    table_name: QualifiedName
    key_name: str | None = None

    KIND: ClassVar[str] = 'set_kms_key_name'



from typing import ClassVar

from liti.core.model.v1.operation.data.base import Operation
from liti.core.model.v1.schema import MaterializedView, QualifiedName, View


class CreateView(Operation):
    """ Semantics: CREATE OR REPLACE """

    view: View

    KIND: ClassVar[str] = 'create_view'


class DropView(Operation):
    """ Semantics: DROP """

    view_name: QualifiedName

    KIND: ClassVar[str] = 'drop_view'


class CreateMaterializedView(Operation):
    """ Semantics: CREATE OR REPLACE """

    materialized_view: MaterializedView

    KIND: ClassVar[str] = 'create_materialized_view'


class DropMaterializedView(Operation):
    """ Semantics: DROP """

    materialized_view_name: QualifiedName

    KIND: ClassVar[str] = 'drop_materialized_view'



from math import isclose
from typing import Any, Callable, ClassVar, Generator, TYPE_CHECKING

from pydantic import BaseModel

from liti.core.context import Context

# avoid circular import errors by delaying the import of model types
# which need to use types from this file
if TYPE_CHECKING:
    from liti.core.model.v1.datatype import Array, BigNumeric, Float, Int, Numeric
    from liti.core.model.v1.schema import MaterializedView, Partitioning, Table, View


class Defaulter:
    """ Observer interface for backends to implement to define defaults

    Default methods update None values to their defaults.
    """

    def defaults_noop(self, node: Any, context: Context):
        pass

    def int_defaults(self, node: 'Int', context: Context):
        pass

    def float_defaults(self, node: 'Float', context: Context):
        pass

    def numeric_defaults(self, node: 'Numeric', context: Context):
        pass

    def big_numeric_defaults(self, node: 'BigNumeric', context: Context):
        pass

    def partitioning_defaults(self, node: 'Partitioning', context: Context):
        pass

    def table_defaults(self, node: 'Table', context: Context):
        pass

    def view_defaults(self, node: 'View', context: Context):
        if node.select_sql is None and node.select_file is not None:
            with open(context.target_dir / node.select_file) as f:
                node.select_sql = f.read()

    def materialized_view_defaults(self, node: 'MaterializedView', context: Context):
        if node.select_sql is None and node.select_file is not None:
            with open(context.target_dir / node.select_file) as f:
                node.select_sql = f.read()


class Defaultable:
    """ Observable interface for the model to implement """

    DEFAULT_METHOD: ClassVar[str] = 'defaults_noop'

    def set_defaults(self, defaulter: Defaulter, context: Context):
        """ Updates the object with defaults applied

        This method should call set_defaults on the object's children.
        """

        getattr(defaulter, self.__class__.DEFAULT_METHOD)(self, context)


class Validator:
    """ Observer interface for backends to implement to validate the model

    Validation methods fix invalid values and raise if still invalid.
    """

    def noop_validate(self, node: Any, context: Context):
        pass

    def validate_int(self, node: 'Int', context: Context):
        pass

    def validate_float(self, node: 'Float', context: Context):
        pass

    def validate_numeric(self, node: 'Numeric', context: Context):
        pass

    def validate_big_numeric(self, node: 'BigNumeric', context: Context):
        pass

    def validate_array(self, node: 'Array', context: Context):
        pass

    def validate_partitioning(self, node: 'Partitioning', context: Context):
        pass

    def validate_view(self, node: 'View', context: Context):
        if not node.select_sql:
            raise ValueError(f'View {node.name} has no select SQL')

    def validate_materialized_view(self, node: 'MaterializedView', context: Context):
        if not node.select_sql:
            raise ValueError(f'Materialized view {node.name} has no select SQL')


class Validatable:
    """ Observable interface for the model to implement """

    VALIDATE_METHOD: ClassVar[str] = 'noop_validate'

    def liti_validate(self, validator: Validator, context: Context):
        """ Raises if not valid

        This method should call liti_validate on the object's children.
        """

        getattr(validator, self.__class__.VALIDATE_METHOD)(self, context)


def is_match(match: Any, value: Any) -> bool:
    # circular import
    from liti.core.model.v1.schema import ValidatedString

    if isinstance(match, dict) and isinstance(value, list | tuple | set):
        # skip checking collection items
        return True
    elif isinstance(value, LitiModel):
        # dig deeper into the model
        return all(is_match(inner, getattr(value, field)) for field, inner in match.items())
    elif isinstance(match, float) and isinstance(value, float):
        return isclose(match, value)
    elif isinstance(value, ValidatedString):
        # avoids having to specify '.string' in templates
        return match == value.string
    else:
        # match must be on the left hand side for STAR comparisons
        return match == value


class Star:
    """ Star is used to match everything """

    def __eq__(self, other):
        return True

    def __getitem__(self, item):
        return self

    def get(self, *args, **kwargs):
        return self

    def items(self):
        return iter(())


STAR = Star()


class LitiModel(BaseModel, Defaultable, Validatable):
    """ Base class for all Liti model classes """

    @classmethod
    def by_name(cls, name: str) -> type['LitiModel']:
        # ensure LitiModel subclasses are imported first
        # noinspection PyUnresolvedReferences
        import liti.core.model.v1.subclasses

        return {
            subclass.__name__: subclass
            for subclass in LitiModel.__subclasses__()
        }[name]

    def set_defaults(self, defaulter: Defaulter, context: Context):
        for field_name in self.__pydantic_fields__.keys():
            field = getattr(self, field_name)

            if isinstance(field, Defaultable):
                field.set_defaults(defaulter, context)

        super().set_defaults(defaulter, context)

    def liti_validate(self, validator: Validator, context: Context):
        for field_name in self.__pydantic_fields__.keys():
            field = getattr(self, field_name)

            if isinstance(field, Validatable):
                field.liti_validate(validator, context)

        super().liti_validate(validator, context)

    def get_roots(self, root: type['LitiModel'], full_match: Any) -> Generator[tuple['LitiModel', Any], None, None]:
        """ Yields all the root nodes of the given type that match the provided `full_match`

        Also yields the remaining local match portion associated with each root in case the template path traverses
        through collection nodes. Those local matches can be used to check each item in the collections.
        """

        # this can be a bit duplicative since each call to `is_match` is recursive,
        # but performance here is not a concern, this is the easiest way to implement
        # the matching logic
        if not is_match(full_match, self):
            return

        if isinstance(self, root):
            yield self, full_match
        else:
            for field_name in self.__pydantic_fields__.keys():
                field = getattr(self, field_name)

                if isinstance(field, list | tuple | set):
                    for item in field:
                        if isinstance(item, LitiModel):
                            yield from item.get_roots(root, full_match[field_name])
                elif isinstance(field, LitiModel):
                    yield from field.get_roots(root, full_match[field_name])

    def get_update_fns(self, path: list[str], matches: list[Any]) -> Generator[Callable[[Any], None], None, None]:
        """ Yields functions to replace selected fields with a provided value

        :param path: a list of field names to recursively traverse through to find the fields to update
        :param matches: a list of either a dict structure of values to compare to the respective fields (functions are
            yielded on equivalence of all fields), or Star to always yield a function, all items in the list must match
        """

        # circular import
        from liti.core.model.v1.schema import ValidatedString

        if not path:
            return

        field_name, *tail = path

        if not hasattr(self, field_name):
            return

        field = getattr(self, field_name)
        field_matches = [m.get(field_name, STAR) for m in matches]

        # stop if any sibling fields do not match
        if not all(
            hasattr(self, f) and is_match(inner, getattr(self, f))
            for m in matches
            for f, inner in m.items()
            if f != field_name
        ):
            return

        if tail:
            # if there are more segments, dig deeper into the model
            if isinstance(field, tuple | list | set):
                # yield for each item that matches in the collection
                for item in field:
                    if isinstance(item, LitiModel):
                        yield from item.get_update_fns(tail, field_matches)
            elif isinstance(field, LitiModel):
                yield from field.get_update_fns(tail, field_matches)
        # yield the leaf field if it matches
        elif all(is_match(fm, field) for fm in field_matches):
            field_type = self.__pydantic_fields__[field_name].annotation

            if issubclass(field_type, ValidatedString):
                # avoids having to specify '.string' in templates
                yield lambda value: setattr(self, field_name, field_type(value))
            else:
                yield lambda value: setattr(self, field_name, value)



from liti.core.context import Context
from liti.core.model.v1.operation.data.table import AddForeignKey, CreateTable, DropConstraint, DropTable, \
    RenameTable, SetClustering, SetDefaultRoundingMode, SetDescription, SetEnableChangeHistory, \
    SetEnableFineGrainedMutations, SetExpirationTimestamp, SetKmsKeyName, SetLabels, SetMaxStaleness, \
    SetPartitionExpiration, SetPrimaryKey, SetRequirePartitionFilter, SetTags
from liti.core.model.v1.operation.ops.base import OperationOps
from liti.core.model.v1.schema import Identifier


class CreateTableOps(OperationOps):
    op: CreateTable

    def __init__(self, op: CreateTable, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.create_table(self.op.table)

    def down(self) -> DropTable:
        return DropTable(table_name=self.op.table.name)

    def is_up(self) -> bool:
        return self.context.db_backend.has_table(self.op.table.name)


class DropTableOps(OperationOps):
    op: DropTable

    def __init__(self, op: DropTable, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.drop_table(self.op.table_name)

    def down(self) -> CreateTable:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return CreateTable(table=sim_table)

    def is_up(self) -> bool:
        return not self.context.db_backend.has_table(self.op.table_name)


class RenameTableOps(OperationOps):
    op: RenameTable

    def __init__(self, op: RenameTable, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.rename_table(self.op.from_name, self.op.to_name)

    def down(self) -> RenameTable:
        return RenameTable(
            from_name=self.op.from_name.with_name(self.op.to_name),
            to_name=self.op.from_name.name,
        )

    def is_up(self) -> bool:
        return self.context.db_backend.has_table(self.op.from_name.with_name(self.op.to_name))


class SetPrimaryKeyOps(OperationOps):
    op: SetPrimaryKey

    def __init__(self, op: SetPrimaryKey, context: Context):
        self.op = op
        self.context = context

    def up(self):
        # TODO: implement removal of the existing primary key if any
        self.context.db_backend.set_primary_key(self.op.table_name, self.op.primary_key)

    def down(self) -> SetPrimaryKey:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return SetPrimaryKey(table_name=self.op.table_name, primary_key=sim_table.primary_key)

    def is_up(self) -> bool:
        return self.context.db_backend.get_table(self.op.table_name).primary_key == self.op.primary_key


class AddForeignKeyOps(OperationOps):
    op: AddForeignKey

    def __init__(self, op: AddForeignKey, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.add_foreign_key(self.op.table_name, self.op.foreign_key)

    def down(self) -> DropConstraint:
        return DropConstraint(table_name=self.op.table_name, constraint_name=Identifier(self.op.foreign_key.name))

    def is_up(self) -> bool:
        fk = self.op.foreign_key
        return self.context.db_backend.get_table(self.op.table_name).foreign_key_map.get(fk.name) == fk


class DropConstraintOps(OperationOps):
    op: DropConstraint

    def __init__(self, op: DropConstraint, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.drop_constraint(self.op.table_name, self.op.constraint_name)

    def down(self) -> AddForeignKey:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)

        return AddForeignKey(
            table_name=self.op.table_name,
            foreign_key=sim_table.foreign_key_map[self.op.constraint_name],
        )

    def is_up(self) -> bool:
        return self.op.constraint_name not in self.context.db_backend.get_table(self.op.table_name).foreign_key_map


class SetPartitionExpirationOps(OperationOps):
    op: SetPartitionExpiration

    def __init__(self, op: SetPartitionExpiration, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_partition_expiration(self.op.table_name, self.op.expiration_days)

    def down(self) -> SetPartitionExpiration:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)

        return SetPartitionExpiration(
            table_name=self.op.table_name,
            expiration_days=sim_table.partitioning and sim_table.partitioning.expiration_days,
        )

    def is_up(self) -> bool:
        partitioning = self.context.db_backend.get_table(self.op.table_name).partitioning
        return (partitioning and partitioning.expiration_days) == self.op.expiration_days


class SetRequirePartitionFilterOps(OperationOps):
    op: SetRequirePartitionFilter

    def __init__(self, op: SetRequirePartitionFilter, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_require_partition_filter(self.op.table_name, self.op.require_filter)

    def down(self) -> SetRequirePartitionFilter:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)

        return SetRequirePartitionFilter(
            table_name=self.op.table_name,
            require_filter=sim_table.partitioning and sim_table.partitioning.require_filter,
        )

    def is_up(self) -> bool:
        partitioning = self.context.db_backend.get_table(self.op.table_name).partitioning
        return (partitioning and partitioning.require_filter) == self.op.require_filter


class SetClusteringOps(OperationOps):
    op: SetClustering

    def __init__(self, op: SetClustering, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_clustering(self.op.table_name, self.op.column_names)

    def down(self) -> SetClustering:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return SetClustering(table_name=self.op.table_name, column_names=sim_table.clustering)

    def is_up(self) -> bool:
        return self.context.db_backend.get_table(self.op.table_name).clustering == self.op.column_names


class SetDescriptionOps(OperationOps):
    op: SetDescription

    def __init__(self, op: SetDescription, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_description(self.op.table_name, self.op.description)

    def down(self) -> SetDescription:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return SetDescription(table_name=self.op.table_name, description=sim_table.description)

    def is_up(self) -> bool:
        return self.context.db_backend.get_table(self.op.table_name).description == self.op.description


class SetLabelsOps(OperationOps):
    op: SetLabels

    def __init__(self, op: SetLabels, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_labels(self.op.table_name, self.op.labels)

    def down(self) -> SetLabels:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return SetLabels(table_name=self.op.table_name, labels=sim_table.labels)

    def is_up(self) -> bool:
        return self.context.db_backend.get_table(self.op.table_name).labels == self.op.labels


class SetTagsOps(OperationOps):
    op: SetTags

    def __init__(self, op: SetTags, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_tags(self.op.table_name, self.op.tags)

    def down(self) -> SetTags:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return SetTags(table_name=self.op.table_name, tags=sim_table.tags)

    def is_up(self) -> bool:
        return self.context.db_backend.get_table(self.op.table_name).tags == self.op.tags


class SetExpirationTimestampOps(OperationOps):
    op: SetExpirationTimestamp

    def __init__(self, op: SetExpirationTimestamp, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_expiration_timestamp(self.op.table_name, self.op.expiration_timestamp)

    def down(self) -> SetExpirationTimestamp:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)

        return SetExpirationTimestamp(
            table_name=self.op.table_name,
            expiration_timestamp=sim_table.expiration_timestamp,
        )

    def is_up(self) -> bool:
        table = self.context.db_backend.get_table(self.op.table_name)
        return table.expiration_timestamp == self.op.expiration_timestamp


class SetDefaultRoundingModeOps(OperationOps):
    op: SetDefaultRoundingMode

    def __init__(self, op: SetDefaultRoundingMode, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_default_rounding_mode(self.op.table_name, self.op.rounding_mode)

    def down(self) -> SetDefaultRoundingMode:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return SetDefaultRoundingMode(table_name=self.op.table_name, rounding_mode=sim_table.default_rounding_mode)

    def is_up(self) -> bool:
        return self.context.db_backend.get_table(self.op.table_name).default_rounding_mode == self.op.rounding_mode


class SetMaxStalenessOps(OperationOps):
    op: SetMaxStaleness

    def __init__(self, op: SetMaxStaleness, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_max_staleness(self.op.table_name, self.op.max_staleness)

    def down(self) -> SetMaxStaleness:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return SetMaxStaleness(table_name=self.op.table_name, max_staleness=sim_table.max_staleness)

    def is_up(self) -> bool:
        return self.context.db_backend.get_table(self.op.table_name).max_staleness == self.op.max_staleness


class SetEnableChangeHistoryOps(OperationOps):
    op: SetEnableChangeHistory

    def __init__(self, op: SetEnableChangeHistory, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_enable_change_history(self.op.table_name, self.op.enabled)

    def down(self) -> SetEnableChangeHistory:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return SetEnableChangeHistory(table_name=self.op.table_name, enabled=sim_table.enable_change_history)

    def is_up(self) -> bool:
        return self.context.db_backend.get_table(self.op.table_name).enable_change_history == self.op.enabled


class SetEnableFineGrainedMutationsOps(OperationOps):
    op: SetEnableFineGrainedMutations

    def __init__(self, op: SetEnableFineGrainedMutations, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_enable_fine_grained_mutations(self.op.table_name, self.op.enabled)

    def down(self) -> SetEnableFineGrainedMutations:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)

        return SetEnableFineGrainedMutations(
            table_name=self.op.table_name,
            enabled=sim_table.enable_fine_grained_mutations,
        )

    def is_up(self) -> bool:
        return self.context.db_backend.get_table(self.op.table_name).enable_fine_grained_mutations == self.op.enabled


class SetKmsKeyNameOps(OperationOps):
    op: SetKmsKeyName

    def __init__(self, op: SetKmsKeyName, context: Context):
        self.op = op
        self.context = context

    def up(self):
        self.context.db_backend.set_kms_key_name(self.op.table_name, self.op.key_name)

    def down(self) -> SetKmsKeyName:
        sim_db = self.simulate(self.context.meta_backend.get_previous_operations())
        sim_table = sim_db.get_table(self.op.table_name)
        return SetKmsKeyName(table_name=self.op.table_name, key_name=sim_table.kms_key_name)

    def is_up(self) -> bool:
        return self.context.db_backend.get_table(self.op.table_name).kms_key_name == self.op.key_name



from abc import ABC, abstractmethod
from pathlib import Path

from liti.core.backend.base import DbBackend, MetaBackend
from liti.core.context import Context
from liti.core.model.v1.operation.data.base import Operation


class OperationOps(ABC):
    op: Operation
    context: Context

    @property
    def db_backend(self) -> DbBackend:
        return self.context.db_backend

    @property
    def meta_backend(self) -> MetaBackend:
        return self.context.meta_backend

    @property
    def target_dir(self) -> Path | None:
        return self.context.target_dir

    @staticmethod
    def simulate(operations: list[Operation]) -> DbBackend:
        # circular imports
        from liti.core.backend.memory import MemoryDbBackend, MemoryMetaBackend
        from liti.core.runner import MigrateRunner

        sim_context = Context(
            db_backend=MemoryDbBackend(),
            meta_backend=MemoryMetaBackend(),
            target_operations=operations,
            silent=True,
        )

        MigrateRunner(context=sim_context).run(wet_run=True)
        return sim_context.db_backend

    @classmethod
    def get_attachment(cls, op: Operation) -> type['OperationOps']:
        # ensure OperationOps subclasses are imported first
        # noinspection PyUnresolvedReferences
        import liti.core.model.v1.operation.ops.subclasses

        return {
            getattr(subclass, '__annotations__')['op']: subclass
            for subclass in OperationOps.__subclasses__()
        }[type(op)]

    @abstractmethod
    def up(self):
        """ Apply the operation """
        pass

    @abstractmethod
    def down(self) -> Operation:
        """ Build the inverse operation """
        pass

    @abstractmethod
    def is_up(self) -> bool:
        """ True if the operation is applied

        If the operation is applied when `is_up` is called, it assumes this is the most recently applied operation.
        Otherwise, the behavior is undefined.
        Can return True even if the metadata is not up to date.
        Useful for recovering from failures that left the migrations in an inconsistent state.
        """
        pass
